/*
   Ray Tracing Multiple Lights Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to build off of the previous ray tracing demos
   and add another light source in the scene.  To do this we will alter the functions
   in RayTracer.cpp to take a pointer to an array of lights instead of just 1 light.
   Also in the LightShadeShape() we will loop through all lights instead of dealing
   with just 1.  Other than that the code is no different from the last demo.
*/


// Demo: Ray Tracing Multiple Lights.


// All the necessary includes...
#include<iostream>
#include"Shape.h"
#include"RayTracer.h"
#include"Light.h"
#include"TgaImage.h"

using namespace std;

// Width/Height resolution and total number of objects.
#define WIDTH_RES 640
#define HEIGHT_RES 480
#define TOTAL_SHAPES 14
#define TOTAL_LIGHTS 2


// Function prototype...
bool InitializeRT();
void ClearBuffer(unsigned char *buffer, int size, int r, int g, int b);
void Shutdown();


// Primary render array and the look up array of directions.
unsigned char *primaryBuffer = 0;
CVector4 *directionList = 0;

// List of objects, light source, and tga object to save the rendered scene.
Shape shapes[TOTAL_SHAPES];
CLight lights[TOTAL_LIGHTS];


int main()
{
   cout << "Ultimate Game Programming Ray Tracer..." << endl;
   cout << "Created by Allen Sherrod..." << endl << endl;

   // Initialize all arrays and data nescessary.  If it fails then we can't
   // perform our ray tracing.
   if(!InitializeRT())
      {
         cout << "Initialization failed!  Exiting." << endl;
         return 0;
      }

   cout << "Clearing primary buffer..." << endl;

   // Clear the rendered scene to a default color (I choose black).
   ClearBuffer(primaryBuffer, WIDTH_RES * HEIGHT_RES * 3, 0, 0, 0);
   
   cout << "Primary buffer cleared and ready to be drawn to..." << endl << endl;
   cout << "Starting trace." << endl;

   // Trace through the scene and create the rendered scene.
	TraceScene(shapes, TOTAL_SHAPES, lights, TOTAL_LIGHTS, primaryBuffer,
	           directionList, WIDTH_RES, HEIGHT_RES);

   cout << "Trace over." << endl << endl;
   
   cout << "Preparing to save scene as RayTracedScene.tga..." << endl;
   
   // Take our ray traced scene and save it as a .tga image file.
   WriteTGA("RayTracedScene.tga", WIDTH_RES, HEIGHT_RES, primaryBuffer);

   cout << "Traced scene saved...shut down starting..." << endl << endl;

   // Release all resources and data we used.
   Shutdown();

   cout << "Exiting..." << endl;

   return 1;
}


bool InitializeRT()
{
   // Release all previous data.
   Shutdown();

   // Allocate rendering buffer.
	primaryBuffer = new unsigned char[WIDTH_RES * HEIGHT_RES * 3];

   // Error checking...
   if(primaryBuffer == NULL)
      return false;

	// Allocate the ray directions.
	directionList = GenerateRayDirections(WIDTH_RES, HEIGHT_RES, 255);

   // Error checking...
   if(directionList == NULL)
      return false;

   // Setup the lights.
   lights[0] = CLight(CVector4(400, 200, -600), 1, 1, 1);
   lights[1] = CLight(CVector4(-400, 200, -600), 1, 1, 1);

   // Initialize the shapes we are using.
   shapes[0].type = SPHERE;
   shapes[0].sphere = new Sphere(-300, -150, 0, 100, 0.5, 0.25, 1);

   shapes[1].type = SPHERE;
   shapes[1].sphere = new Sphere(300, -150, 0, 100, 1, 0.5, 0.5);
   
   // Ground.
   shapes[2].type = TRIANGLE;
   shapes[2].triangle = new Triangle(CVector4(-600, -300, -800),
                                     CVector4(-600, -300, 800),
                                     CVector4(600, -300, 800), 1, 1, 1);
   shapes[3].type = TRIANGLE;
   shapes[3].triangle = new Triangle(CVector4(600, -300, 800),
                                     CVector4(600, -300, -800),
                                     CVector4(-600, -300, -800), 1, 1, 1);
   // Left.
   shapes[4].type = TRIANGLE;
   shapes[4].triangle = new Triangle(CVector4(-600, -300, 800),
                                     CVector4(-600, -300, -800),
                                     CVector4(-600, 550, -800), 1, 0, 0);
   shapes[5].type = TRIANGLE;
   shapes[5].triangle = new Triangle(CVector4(-600, 550, -800),
                                     CVector4(-600, 550, 800),
                                     CVector4(-600, -300, 800), 1, 0, 0);
   // Right.
   shapes[6].type = TRIANGLE;
   shapes[6].triangle = new Triangle(CVector4(600, -300, -800),
                                     CVector4(600, -300, 800),
                                     CVector4(600, 550, 800), 0, 0, 1);
   shapes[7].type = TRIANGLE;
   shapes[7].triangle = new Triangle(CVector4(600, 550, 800),
                                     CVector4(600, 550, -800),
                                     CVector4(600, -300, -800), 0, 0, 1);
   // Roof.
   shapes[8].type = TRIANGLE;
   shapes[8].triangle = new Triangle(CVector4(-600, 550, 800),
                                     CVector4(-600, 550, -800),
                                     CVector4(600, 550, -800), 1, 1, 1);
   shapes[9].type = TRIANGLE;
   shapes[9].triangle = new Triangle(CVector4(600, 550, -800),
                                     CVector4(600, 550, 800),
                                     CVector4(-600, 550, 800), 1, 1, 1);
   // Front.
   shapes[10].type = TRIANGLE;
   shapes[10].triangle = new Triangle(CVector4(-600, -300, -800),
                                     CVector4(600, -300, -800),
                                     CVector4(600, 550, -800), 1, 1, 1);
   shapes[11].type = TRIANGLE;
   shapes[11].triangle = new Triangle(CVector4(600, 550, -800),
                                     CVector4(-600, 550, -800),
                                     CVector4(-600, -300, -800), 1, 1, 1);
   // Back.
   shapes[12].type = TRIANGLE;
   shapes[12].triangle = new Triangle(CVector4(600, -300, 800),
                                     CVector4(-600, -300, 800),
                                     CVector4(-600, 550, 800), 0, 1, 0);
   shapes[13].type = TRIANGLE;
   shapes[13].triangle = new Triangle(CVector4(-600, 550, 800),
                                     CVector4(600, 550, 800),
                                     CVector4(600, -300, 800), 0, 1, 0);

   return true;
}


void ClearBuffer(unsigned char *buffer, int size, int r, int g, int b)
{
   // Loop through the entire array and set the buffer to the specified color.
   for(int p = 0; p < size; p += 3)
      {
         buffer[p + 0] = r; buffer[p + 1] = g; buffer[p + 2] = b;
      }
}


void Shutdown()
{
   // Release all resources we used.
   if(directionList != NULL) delete[] directionList;
   if(primaryBuffer != NULL) delete[] primaryBuffer;


   for(int i = 0; i < TOTAL_SHAPES; i++)
      {
         if(shapes[i].type == SPHERE && shapes[i].sphere != NULL)
            delete shapes[i].sphere;
         else if(shapes[i].type == TRIANGLE && shapes[i].triangle != NULL)
            delete shapes[i].triangle;
         else if(shapes[i].type == PLANE && shapes[i].plane != NULL)
            delete shapes[i].plane;
      }
}


// Recap:

// Now we can have as many light sources, triangles, planes, and sphere as we
// want in the scene.  There are a few things we did not take into consideration
// like point lights, spot lights, materials (kinda did but not really), specular
// lighting, emissive, and light intensity.  But I am sure you can add all of this
// if you wanted to, for our proposes we only need global diffuse light.


// Copyright February 2005
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com